
class ConnectionInterrupted(Exception):
    """连接中断异常"""
    def __str__(self):
        error_type = self.__class__.__name__
        error_msg = "An error occurred while connecting to redis cluster"

        return "Redis Cluster %s: %s" % (error_type, error_msg)


class CompressorError(Exception):
    pass




from __future__ import absolute_import

from lz4.frame import compress as _compress, decompress as _decompress

from ..exceptions import CompressorError
from .base import BaseCompressor


class Lz4Compressor(BaseCompressor):
    min_length = 15

    def compress(self, value):
        if len(value) > self.min_length:
            return _compress(value)
        return value

    def decompress(self, value):
        try:
            return _decompress(value)
        except Exception as e:
            raise CompressorError(e)



import functools
import logging

from django.conf import settings
from django.core.cache.backends.base import BaseCache

from .utils import load_class
from .exceptions import ConnectionInterrupted

DEDIS_CLUSTER_IGNORE_EXCEPTIONS = getattr(settings, 'DEDIS_CLUSTER_IGNORE_EXCEPTIONS', False)
DEDIS_CLUSTER_LOG_IGNORED_EXCEPTIONS = getattr(settings, 'DEDIS_CLUSTER_LOG_IGNORED_EXCEPTIONS', False)
DEDIS_CLUSTER_LOGGER = getattr(settings, 'DEDIS_CLUSTER_LOGGER', False)

if DEDIS_CLUSTER_IGNORE_EXCEPTIONS:
    logger = logging.getLogger((DEDIS_CLUSTER_LOGGER or __name__))


def omit_exception(method=None, return_value=None):
    """连接异常处理修饰器"""
    
    if method is None:
        return functools.partial(omit_exception, return_value=return_value)
    
    @functools.wraps(method)
    def _decorator(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except ConnectionInterrupted as e:
            if self._ignore_exceptions:
                if DEDIS_CLUSTER_LOG_IGNORED_EXCEPTIONS:
                    logger.error(str(e))
                    
                return return_value
            raise e
    return _decorator


class RedisClusterCache(BaseCache):
    def __init__(self, server, params):
        super(RedisClusterCache, self).__init__(params)
        self._server = server
        self._params = params
        options = params.get('OPTIONS', {})
        self._client_cls = options.get("CLIENT_CLASS", "dedis_cluster.client.DefaultClient")
        self._client_cls = load_class(self._client_cls)
        self._client = None
        self._ignore_exceptions = options.get('IGNORE_EXCEPTIONS', DEDIS_CLUSTER_IGNORE_EXCEPTIONS)

    @property
    def client(self):
        """Lazy client connection property"""
        if self._client is None:
            self._client = self._client_cls(self._server, self._params, self)
        return self._client

    @omit_exception
    def add(self, *args, **kwargs):
        return self.client.add(*args, **kwargs)

    @omit_exception
    def set(self, *args, **kwargs):
        return self.client.set(*args, **kwargs)

    @omit_exception
    def set_many(self, *args, **kwargs):
        return self.client.set_many(*args, **kwargs)
    
    @omit_exception
    def get(self, key, default=None, version=None):
        try:
            return self.client.get(key, default=default, version=version)
        except ConnectionInterrupted as e:
            if DEDIS_CLUSTER_LOG_IGNORED_EXCEPTIONS:
                logger.error(str(e))
            return default

    @omit_exception(return_value={})
    def get_many(self, *args, **kwargs):
        return self.client.get_many(*args, **kwargs)
    
    @omit_exception
    def delete(self, *args, **kwargs):
        return self.client.delete(*args, **kwargs)

    @omit_exception
    def delete_pattern(self, *args, **kwargs):
        return self.client.delete_pattern(*args, **kwargs)

    @omit_exception
    def delete_many(self, *args, **kwargs):
        return self.client.delete_many(*args, **kwargs)

    @omit_exception
    def clear(self):
        return self.client.clear()
    
    @omit_exception
    def keys(self, *args, **kwargs):
        return self.client.keys(*args, **kwargs)
    
    @omit_exception
    def has_key(self, *args, **kwargs):
        return self.client.has_key(*args, **kwargs)

    @omit_exception
    def lock(self, *args, **kwargs):
        return self.client.lock(*args, **kwargs)

    @omit_exception
    def incr_version(self, *args, **kwargs):
        return self.client.incr_version(*args, **kwargs)
    
    @omit_exception
    def incr(self, *args, **kwargs):
        return self.client.incr(*args, **kwargs)

    @omit_exception
    def decr(self, *args, **kwargs):
        return self.client.decr(*args, **kwargs)
    
    @omit_exception
    def iter_keys(self, *args, **kwargs):
        return self.client.iter_keys(*args, **kwargs)

    @omit_exception
    def ttl(self, *args, **kwargs):
        return self.client.ttl(*args, **kwargs)

    @omit_exception
    def persist(self, *args, **kwargs):
        return self.client.persist(*args, **kwargs)

    @omit_exception
    def expire(self, *args, **kwargs):
        return self.client.expire(*args, **kwargs)




from __future__ import absolute_import, unicode_literals

import json
from django.core.serializers.json import DjangoJSONEncoder
try:
    from django.utils.encoding import force_bytes
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import smart_bytes as force_bytes
    from django.utils.encoding import force_unicode as force_text


from .base import BaseSerializer


class JSONSerializer(BaseSerializer):
    def dumps(self, value):
        return json.dumps(value, cls=DjangoJSONEncoder).encode()
        # return force_bytes(json.dumps(value))

    def loads(self, value):
        return json.loads(value.decode())
        # return json.loads(force_text(value))



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import lzma

from ..exceptions import CompressorError
from .base import BaseCompressor


class LzmaCompressor(BaseCompressor):
    min_length = 100
    preset = 4

    def compress(self, value):
        if len(value) > self.min_length:
            return lzma.compress(value, preset=self.preset)
        return value

    def decompress(self, value):
        try:
            return lzma.decompress(value)
        except lzma.LZMAError as e:
            raise CompressorError(e)




from __future__ import absolute_import, unicode_literals

try:
    import cPickle as pickle
except ImportError:
    import pickle

from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import force_bytes

from .base import BaseSerializer


class PickleSerializer(BaseSerializer):
    def __init__(self, options):
        self._pickle_version = -1
        self.setup_pickle_version(options)
        super(PickleSerializer, self).__init__(options)

    def setup_pickle_version(self, options):
        if "PICKLE_VERSION" in options:
            try:
                self._pickle_version = int(options["PICKLE_VERSION"])
            except (ValueError, TypeError):
                raise ImproperlyConfigured("PICKLE_VERSION value must be an integer")

    def dumps(self, value):
        return pickle.dumps(value, self._pickle_version)

    def loads(self, value):
        return pickle.loads(force_bytes(value))



from importlib import import_module

from django.utils.encoding import smart_text
from django.core.exceptions import ImproperlyConfigured
integer_types = (int,)
    

class CacheKey(object):
    """
    A stub string class that we can use to check if a key was created already.
    """
    def __init__(self, key):
        self._key = key

    def __str__(self):
        return smart_text(self._key)

    def original_key(self):
        key = self._key.rsplit(":", 1)[1]
        return key
    

def load_class(path):
    """通过路径加载类"""

    mod_name, klass_name = path.rsplit('.', 1)

    try:
        mod = import_module(mod_name)
    except AttributeError as e:
        raise ImproperlyConfigured('Error importing {0}: "{1}"'.format(mod_name, e))

    try:
        klass = getattr(mod, klass_name)
    except AttributeError:
        raise ImproperlyConfigured('Module "{0}" does not define a "{1}" class'.format(mod_name, klass_name))

    return klass


def default_reverse_key(key):
    return key.split(':', 2)[2]





from __future__ import absolute_import

import zlib

from ..exceptions import CompressorError
from .base import BaseCompressor


class ZlibCompressor(BaseCompressor):
    min_length = 15
    preset = 6

    def compress(self, value):
        if len(value) > self.min_length:
            return zlib.compress(value, self.preset)
        return value

    def decompress(self, value):
        try:
            return zlib.decompress(value)
        except zlib.error as e:
            raise CompressorError(e)



from redis._compat import urlparse

from rediscluster import StrictRedisCluster
from rediscluster.connection import ClusterConnectionPool


class ConnectionFactory(object):

    def __init__(self, server, options):
        self._server = server
        self._options = options
    
    @property
    def connection(self):
        """创建连接"""
        params = self._make_connection_params()
        print('params: ', params)
        pool = self._get_connection_pool(params)
        connection = StrictRedisCluster(connection_pool=pool)
        return connection

    def _make_connection_params(self):
        """生成连接需要的参数"""
        kwargs = {}
        
        socket_timeout = self._options.get("SOCKET_TIMEOUT", None)
        if socket_timeout:
            assert isinstance(socket_timeout, (int, float)), \
                "Socket timeout should be float or integer"
            kwargs["socket_timeout"] = socket_timeout
        
        socket_connect_timeout = self._options.get("SOCKET_CONNECT_TIMEOUT", None)
        if socket_connect_timeout:
            assert isinstance(socket_connect_timeout, (int, float)), \
                "Socket connect timeout should be float or integer"
            kwargs["socket_connect_timeout"] = socket_connect_timeout
            
        return kwargs
            
    def _get_connection_pool(self, params):
        """创建连接池"""
        cp_params = dict(params)
        # cp_params.update({'decode_responses': True})
        startup_nodes = self._parse_startup_nodes()
        pool = ClusterConnectionPool(startup_nodes, **cp_params)
        return pool

    def _parse_startup_nodes(self):
        """解析所有服务节点"""
        startup_nodes = []
        
        for url_string in self._server:
            url = urlparse(url_string)
            if url.hostname:
                startup_nodes.append({'host': url.hostname, 'port': int(url.port or 6379)})
                
        return startup_nodes
    
    
def get_connection_factory(server, options=None):
    
    return ConnectionFactory(server, options)





from __future__ import absolute_import, unicode_literals

import msgpack

from .base import BaseSerializer


class MSGPackSerializer(BaseSerializer):
    def dumps(self, value):
        return msgpack.dumps(value)

    def loads(self, value):
        return msgpack.loads(value, encoding="utf-8")




import warnings
import socket
from collections import OrderedDict

from django.utils.encoding import smart_text

from django.core.cache.backends.base import get_key_func
from django.core.exceptions import ImproperlyConfigured

try:
    from django.core.cache.backends.base import DEFAULT_TIMEOUT
except ImportError:
    DEFAULT_TIMEOUT = object()
    
try:
    from redis.exceptions import TimeoutError, ResponseError
    _main_exceptions = (TimeoutError, ResponseError, ConnectionError, socket.timeout)
except ImportError:
    _main_exceptions = (ConnectionError, socket.timeout)
from ..exceptions import CompressorError
from six import integer_types
from ..utils import CacheKey, load_class
from ..exceptions import ConnectionInterrupted
from .. import connection


class DefaultClient(object):
    
    def __init__(self, server, params, backend):
        self._server = server
        self._params = params
        self._backend = backend
        
        self.reverse_key = get_key_func(params.get('REVERS_KEY_FUNCTION') or
                                        'dedis_cluster.utils.default_reverse_key')
        
        if not self._server:
            raise ImproperlyConfigured('Missing connections string')
        
        # 初始空clients
        if not isinstance(self._server, (list, tuple, set)):
            self._server = self._server.split(',')
        
        self._options = params.get('OPTIONS', {})
        serializer_path = self._options.get("SERIALIZER", "dedis_cluster.serializers.pickle.PickleSerializer")
        serializer_cls = load_class(serializer_path)
        compressor_path = self._options.get("COMPRESSOR", "dedis_cluster.compressors.identity.IdentityCompressor")
        compressor_cls = load_class(compressor_path)
        self._serializer = serializer_cls(options=self._options)
        self._compressor = compressor_cls(options=self._options)
        self._connect_factory = connection.get_connection_factory(server=self._server, options=self._options)
        self._client = self._connect_factory.connection

    def __contains__(self, key):
        return self.has_key(key)

    def add(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        """添加一个键值，如果存在则返回失败"""
        return self.set(key, value, timeout, version=version, nx=True)
    
    def set(self, key, value, timeout=DEFAULT_TIMEOUT, version=None, nx=False, xx=False):
        """缓存一个键值"""
        nkey = self._make_key(key, version=version)
        nvalue = self.encode(value)

        if timeout is True:
            warnings.warn("Using True as timeout value, is now deprecated.", DeprecationWarning)
            timeout = self._backend.default_timeout
        
        if timeout == DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout
        
        try:
            if timeout is not None:
                if timeout > 0:
                    timeout = int(timeout)
                elif timeout <= 0:
                    if nx:
                        timeout = None
                    else:
                        return self.delete(key, version=version)
            return self._client.set(nkey, nvalue, nx=nx, ex=timeout, xx=xx)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)

    def set_many(self, data, timeout=DEFAULT_TIMEOUT, version=None):
        """设置多个缓存"""
        try:
            pipeline = self._client.pipeline()
            for key, value in data.items():
                self.set(key, value, timeout, version=version)
            pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)

    def get(self, key, default=None, version=None):
        """如果key存在时取回键值，否则返回默认值"""
        key = self._make_key(key, version=version)
        try:
            print(type(self._client))

            value = self._client.get(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)
        
        if value is None:
            return default
        
        return self.decode(value)

    def get_many(self, keys, version=None):
        """查询多个key缓存"""
        if not keys:
            return {}
        recovered_data = OrderedDict()
        new_keys = [self._make_key(k, version=version) for k in keys]
        map_keys = dict(zip(new_keys, keys))
        try:
            results = self._client.mget(*new_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)
        for key, value in zip(new_keys, results):
            if value is None:
                continue
            recovered_data[map_keys[key]] = self.decode(value)
        return recovered_data

    def keys(self, search, version=None):
        """查询缓存的keys"""
        pattern = self._make_key(search, version=version)
        try:
            encoding_map = [smart_text(k) for k in self._client.keys(pattern)]
            return [self.reverse_key(k) for k in encoding_map]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)

    def delete(self, key, version=None):
        """删除key的缓存"""
        try:
            self._client.delete(self._make_key(key, version=version))
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)

    def delete_pattern(self, pattern, version=None):
        """删除所有符合条件的key"""
        pattern = self._make_key(pattern, version=version)
        try:
            count = 0
            for key in self._client.scan_iter(pattern):
                self._client.delete(key)
                count += 1
            return count
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)

    def delete_many(self, keys, version=None):
        """一次性删除多个key"""
        keys = [self._make_key(k, version=version) for k in keys]
        if not keys:
            return 
        try:
            return self._client.delete(*keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)

    def has_key(self, key, version=None):
        """是否存在key"""
        key = self._make_key(key, version=version)
        try:
            return self._client.exists(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)

    def clear(self):
        """清空所有缓存"""
        self._client.flushdb()

    def lock(self, key, version=None, timeout=None, sleep=0.1, blocking_timeout=None):
        key = self._make_key(key, version)
        return self._client.lock(key, timeout=timeout, sleep=sleep, blocking_timeout=blocking_timeout)
    
    def decode(self, value):
        """
        Decode the given value.
        """
        try:
            value = int(value)
        except (ValueError, TypeError):
            try:
                value = self._compressor.decompress(value)
            except CompressorError:
                # Handle little values, chosen to be not compressed
                pass
            value = self._serializer.loads(value)
        return value

    def encode(self, value):
        """
        Encode the given value.
        """
        if isinstance(value, bool) or not isinstance(value, integer_types):
            value = self._serializer.dumps(value)
            value = self._compressor.compress(value)
            return value

        return value
    
    def incr_version(self, key, delta=1, version=None):
        """
        Adds delta to the cache version for the supplied key. Returns the
        new version.
        """
        if version is None:
            version = self._backend.version
            
        old_key = self._make_key(key, version)
        value = self.get(old_key, version=version)
        
        try:
            ttl = self._client.ttl(old_key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)
        
        if value is None:
            raise ValueError("Key '%s' not found" % key)
    
        if isinstance(key, CacheKey):
            new_key = self._make_key(key.original_key(), version=version + delta)
        else:
            new_key = self._make_key(key, version=version + delta)
            
        self.set(new_key, value, timeout=ttl)
        self.delete(old_key)
        return version + delta
    
    def persist(self, key, version=None):
        key = self._make_key(key, version=version)
        if self._client.exists(key):
            self._client.persist(key)

    def expire(self, key, timeout, version=None):
        key = self._make_key(key, version=version)
        if self._client.exists(key):
            self._client.expire(key, timeout)
            
    def _incr(self, key, delta=1, version=None):
        key = self._make_key(key, version=version)
        try:
            if not self._client.exists(key):
                raise ValueError("Key '%s' not found" % key)
            try:
                value = self._client.incr(key, delta)
            except ResponseError:
                timeout = self._client.ttl(key)
                value = self.get(key, version=version) + delta
                self.set(key, value, version=version, timeout=timeout)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=self._client, parent=e)
        return value

    def incr(self, key, delta=1, version=None):
        """
        Add delta to value in the cache. If the key does not exist, raise a
        ValueError exception.
        """
        return self._incr(key=key, delta=delta, version=version)

    def decr(self, key, delta=1, version=None):
        """
        Decreace delta to value in the cache. If the key does not exist, raise a
        ValueError exception.
        """
        return self._incr(key=key, delta=-delta, version=version)
    
    def ttl(self, key, version=None):
        """
        Executes TTL redis command and return the "time-to-live" of specified key.
        If key is a non volatile key, it returns None.
        """
        key = self._make_key(key, version=version)
        if not self._client.exists(key):
            return 0
        
        t = self._client.ttl(key)
        return t >= 0 and t or None

    def iter_keys(self, search, itersize=None, version=None):
        """
        Same as keys, but uses redis >= 2.8 cursors
        for make memory efficient keys iteration.
        """
        pattern = self._make_key(search, version=version)
        for item in self._client.scan_iter(match=pattern, count=itersize):
            item = smart_text(item)
            yield self.reverse_key(item)
    
    def _make_key(self, key, version=None):
        """格式化key"""
        if isinstance(key, CacheKey):
            return key
        return CacheKey(self._backend.make_key(key, version))





class BaseCompressor(object):
    def __init__(self, options):
        self._options = options

    def compress(self, value):
        raise NotImplementedError

    def decompress(self, value):
        raise NotImplementedError



from setuptools import setup

description = """
Redis-cluster plugin for Django
"""

setup(
    name="dedis-cluster",
    url="https://github.com/feiwencaho/dedis-cluster.git",
    author="feivincent",
    author_email="hi@feiwenchao.me",
    version="1.0.1",
    packages=[
        "dedis_cluster",
        "dedis_cluster.client",
        "dedis_cluster.serializers",
        "dedis_cluster.compressors",
    ],
    description=description.strip(),
    install_requires=[
        "Django>=1.9.6",
        "redis>=2.10.5",
        "redis-py-cluster>=1.2.0",
        "msgpack-python>=0.4.7",
    ],
    zip_safe=False,
    include_package_data=True,
    package_data={
        "": ["*.html"],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "Environment :: Web Environment",
        "Framework :: Django",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
    ],
)




from django.core.cache import caches


def get_cache(alias):
    
    return caches[alias]


def get_redis_connection(alias='default', write=True):
    """
    Helper used for obtain a raw redis client.
    """

    cache = get_cache(alias)
    if not hasattr(cache.client, 'get_client'):
        raise NotImplementedError("This backend does not supports this feature")

    return cache.client.get_client(write)




from __future__ import absolute_import, unicode_literals


class BaseSerializer(object):
    def __init__(self, options):
        pass

    def dumps(self, value):
        raise NotImplementedError

    def loads(self, value):
        raise NotImplementedError
